### Cross Validation Analyses
### CV has two parts, each applied to both primary and secondary, sensitivity, datasets.
### First, we follow typical CV procedure with k-fold CV with k=10.
### Second, we use forward chaining CV to grapple with the temporal autocorrelation in these data.



library(mgcv)
library(lme4)
library(gamm4)
library(car)
library(caret)
library(optimx)
library(ggplot2)
library(plyr)
library(dplyr)



### 1. Standard k-fold cross validation


### Age only (Differing rank initialization)
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$chimpcode)
table(df_init_allNew[cvInd$Fold02,]$chimpcode) # check

## Cross validating 
xy = df_init_allNew
cv_init_age <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_age[k] <- mean((y - as.vector(pred$fit))^2)
}
cv_mean_init_age <- mean(cv_init_age)

mse_init_age = deviance(bam.card.init.cs.OnlyAge)/nobs(bam.card.init.cs.OnlyAge)


### Age + D
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$Chimp_ID)
table(df_init_allNew[cvInd$Fold02,]$Chimp_ID) # check

## Cross validating 
xy = df_init_allNew
cv_init_ageD <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(age_cz, D_new_resid, bs='cs')#, k=15) # what is a good k??? 
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_ageD[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_mean_init_ageD <- mean(cv_init_ageD)

mse_init_ageD = deviance(bam.card.init.cs.AgeDnr)/nobs(bam.card.init.cs.AgeDnr)


## Age + C
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$Chimp_ID)
table(df_init_allNew[cvInd$Fold02,]$Chimp_ID) # check

## Cross validating 
xy = df_init_allNew
cv_init_ageC <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(C_resid, bs='cs') 
    + ti(age_cz, C_resid, bs='cs')#, k=15) # what is a good k??? 
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, C_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_ageC[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_mean_init_ageC <- mean(cv_init_ageC)

mse_init_ageC = deviance(bam.card.init.cs.AgeCr)/nobs(bam.card.init.cs.AgeCr)


## Age + D + C
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$Chimp_ID)
table(df_init_allNew[cvInd$Fold02,]$Chimp_ID) # check

## Cross validating 
xy = df_init_allNew
cv_init_ageDC <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(C_resid, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')#, k=15) # what is a good k??? 
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_ageDC[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_mean_init_ageDC <- mean(cv_init_ageDC)

mse_init_ageDC = deviance(bam.card.init.cs.AgeDnrCr)/nobs(bam.card.init.cs.AgeDnrCr)


## Age + all 6
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$Chimp_ID)
table(df_init_allNew[cvInd$Fold02,]$Chimp_ID) # check

## Cross validating 
xy = df_init_allNew
cv_init_ageAll6 <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(C_resid, bs='cs')
    + ti(Ext_cz, bs='cs') + ti(Agr_cz, bs='cs')
    + ti(Opn_cz, bs='cs') + ti(Neu_cz, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')
    + ti(age_cz, Ext_cz, bs='cs') + ti(age_cz, Agr_cz, bs='cs')
    + ti(age_cz, Opn_cz, bs='cs') + ti(age_cz, Neu_cz, bs='cs')
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    + s(chimpcode, Ext_cz, bs='re') + s(chimpcode, Agr_cz, bs='re')
    + s(chimpcode, Opn_cz, bs='re') + s(chimpcode, Neu_cz, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_ageAll6[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_mean_init_ageAll6 <- mean(cv_init_ageAll6)

mse_init_ageAll6 = deviance(bam.card.init.cs.all6)/nobs(bam.card.init.cs.all6)


## Age + all 6 + date RE
set.seed(1234567)
n_folds=10
cvInd=createFolds(df_init_allNew$chimpcode, k=n_folds, list=TRUE)

## Checking stratification
table(df_init_allNew[cvInd$Fold01,]$Chimp_ID)
table(df_init_allNew[cvInd$Fold02,]$Chimp_ID) # check

## CV
xy = df_init_allNew
cv_init_ageWithDate <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  test_i = unlist(cvInd[k])
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  y <- train_xy$EloCardinal
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(C_resid, bs='cs')
    + ti(Ext_cz, bs='cs') + ti(Agr_cz, bs='cs')
    + ti(Opn_cz, bs='cs') + ti(Neu_cz, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')
    + ti(age_cz, Ext_cz, bs='cs') + ti(age_cz, Agr_cz, bs='cs')
    + ti(age_cz, Opn_cz, bs='cs') + ti(age_cz, Neu_cz, bs='cs')
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    + s(chimpcode, Ext_cz, bs='re') + s(chimpcode, Agr_cz, bs='re')
    + s(chimpcode, Opn_cz, bs='re') + s(chimpcode, Neu_cz, bs='re')
    + s(as.numeric(date), bs='cs')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_init_ageWithDate[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_mean_init_ageWithDate <- mean(cv_init_ageWithDate)

mse_init_ageWithDate = deviance(bam.card.init.cs.allPlusDateRE)/nobs(bam.card.init.cs.allPlusDateRE)



### 2. Forward chaining (Time series) approach


## Differing rank initializations
totalNs = table(df_init_allNew$chimpcode)

## Initialize a df with the indices we need:
df.tsFoldPoints = data.frame(chimp=character(28),totalNs=numeric(28),
                        train1=numeric(28),train2=numeric(28),train3=numeric(28),train4=numeric(28),train5=numeric(28),
                        train6=numeric(28),train7=numeric(28),train8=numeric(28),train9=numeric(28),train10=numeric(28),
                        train11=numeric(28) # the last point is needed to close the final for-loop (see below)
                        )
## The above makes a list of end points that we can use to make sequential overlapping sequences 
## to define the training and test sets.


## Fill out the df:
df.tsFoldPoints$totalNs=table(df_init_allNew$chimpcode)
df.tsFoldPoints$chimp=names(table(df_init_allNew$chimpcode))

number_of_chimps = dim(df.tsFoldPoints)[1] # 28

for (i in 1:number_of_chimps){
  ## create a sequence of points for each individual
  df.tsFoldPoints[i,3:13]=seq(from=1, to=df.tsFoldPoints$totalNs[i],length.out=12)[2:12]
}

df.tsFoldPoints[,3:13]=ceiling(df.tsFoldPoints[,3:13])  ## rounding up
  

## Data.frame must be in chronological order:
df_init_allNew = df_init_allNew[order(df_init_allNew$age, df_init_allNew$chimpcode),]
rownames(df_init_allNew) <- seq(length=nrow(df_init_allNew)) 
plot(df_init_allNew$age[df_init_allNew$chimpcode=='FND']) # Test 1
plot(df_init_allNew$age[df_init_allNew$chimpcode=='AO'])  # Test 2
## Looks good.

  
## Make the actual sequences for each chimp, nested within each 'fold'
obs = dim(df_init_allNew)[1]
df.tsTrain=data.frame(chimpcode=character(obs), train01=logical(obs), train02=logical(obs), train03=logical(obs), 
                      train04=logical(obs), train05=logical(obs), train06=logical(obs), train07=logical(obs),
                      train08=logical(obs), train09=logical(obs), train10=logical(obs))
df.tsTest=data.frame(chimpcode=character(obs), test01=logical(obs), test02=logical(obs), test03=logical(obs), 
                     test04=logical(obs), test05=logical(obs), test06=logical(obs), test07=logical(obs),
                     test08=logical(obs), test09=logical(obs), test10=logical(obs))
df.tsTest$Chimp_ID=df_init_allNew$chimpcode
df.tsTrain$Chimp_ID=df_init_allNew$chimpcode

trainIndx=logical(obs)
testIndx=logical(obs)

for (i in 1:n_folds){
  for (j in 1:number_of_chimps){
    chimpIndx=which(df_init_allNew$chimpcode %in% df.tsFoldPoints$chimp[j]) # id indxs for that chimp
    trainWhich=chimpIndx[seq(from=1, to=df.tsFoldPoints[j,i+2], by=1)] # get the training indxs
    testWhich=chimpIndx[seq(from=df.tsFoldPoints[j,i+2]+1, to=df.tsFoldPoints[j,i+3], by=1)] # get the test indxs
    for (k in 1:dim(df_init_allNew)[1]){ # convert indxs numbers to logical vector
      trainIndx[k]=(k %in% trainWhich) 
    }
    for (k in 1:dim(df_init_allNew)[1]){ # convert indxs numbers to logical vector
      testIndx[k]=(k %in% testWhich)
    }
    df.tsTrain[,i+1] = df.tsTrain[i+1] | trainIndx
    df.tsTest[,i+1] = df.tsTest[i+1] | testIndx
  }
}


## Check stratification:
table(df.tsTrain$chimpcode[df.tsTrain$train01])
table(df.tsTrain$chimpcode[df.tsTrain$train02])
table(df.tsTrain$chimpcode[df.tsTrain$train10])
table(df.tsTest$chimpcode[df.tsTest$test01])
table(df.tsTest$chimpcode[df.tsTest$test02])
table(df.tsTest$chimpcode[df.tsTest$test10])
## Training sets grow longer by the same amount each time, and the test set remains the same length.
## This is as expected.


## Cross validate across the sets:

## Age only
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_age <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_age[k] <- mean((y - as.vector(pred$fit))^2)
}
cv_ts_age_avg <- mean(cv_ts_age)


## Dominance only
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_Dnr <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
      + ti(age_cz, D_new_resid, bs='cs')
      + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
      + s(chimpcode, D_new_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_Dnr[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_ts_Dnr_avg <- mean(cv_ts_Dnr)
range(cv_ts_Dnr)


## Conscientiousness only
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_Cr <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
     ti(C_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')#, k=15) # what is a good k??? 
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, C_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_Cr[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_ts_Cr_avg <- mean(cv_ts_Cr)
range(cv_ts_Cr)


## D + C
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_DnrCr <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(C_resid, bs='cs')
    + ti(D_new_resid, bs='cs') 
    + ti(age_cz, C_resid, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_DnrCr[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_ts_DnrCr_avg <- mean(cv_ts_DnrCr)
range(cv_ts_DnrCr)


### All 6 dimensions
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_all6 <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(C_resid, bs='cs')
    + ti(Ext_cz, bs='cs') + ti(Agr_cz, bs='cs')
    + ti(Opn_cz, bs='cs') + ti(Neu_cz, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')
    + ti(age_cz, Ext_cz, bs='cs') + ti(age_cz, Agr_cz, bs='cs')
    + ti(age_cz, Opn_cz, bs='cs') + ti(age_cz, Neu_cz, bs='cs')
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    + s(chimpcode, Ext_cz, bs='re') + s(chimpcode, Agr_cz, bs='re')
    + s(chimpcode, Opn_cz, bs='re') + s(chimpcode, Neu_cz, bs='re') 
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_all6[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_ts_all6_avg <- mean(cv_ts_all6)
range(cv_ts_all6)


### All 6 + date RE
set.seed(1234567) # This should not strictly be necessary because we have already created the 'folds'.
xy = df_init_allNew
cv_ts_plusDateRE <- matrix(NA, nrow = n_folds, ncol = 1)
for (k in 1:n_folds) {
  train_i = as.vector(unlist(df.tsTrain[,k+1]))
  train_xy <- xy[train_i, ]
  test_i = as.vector(unlist(df.tsTest[,k+1]))
  test_xy <- xy[test_i, ]
  fitted_model = bam(
    EloCardinal ~ ti(age_cz, bs='cs', k=7) + 
      ti(D_new_resid, bs='cs') 
    + ti(C_resid, bs='cs')
    + ti(Ext_cz, bs='cs') + ti(Agr_cz, bs='cs')
    + ti(Opn_cz, bs='cs') + ti(Neu_cz, bs='cs')
    + ti(age_cz, D_new_resid, bs='cs')
    + ti(age_cz, C_resid, bs='cs')
    + ti(age_cz, Ext_cz, bs='cs') + ti(age_cz, Agr_cz, bs='cs')
    + ti(age_cz, Opn_cz, bs='cs') + ti(age_cz, Neu_cz, bs='cs')
    + s(chimpcode, bs='re') + s(chimpcode, age_cz, bs='re') # random intercept and slope
    + s(chimpcode, D_new_resid, bs='re')
    + s(chimpcode, C_resid, bs='re')
    + s(chimpcode, Ext_cz, bs='re') + s(chimpcode, Agr_cz, bs='re')
    + s(chimpcode, Opn_cz, bs='re') + s(chimpcode, Neu_cz, bs='re') 
    + s(as.numeric(date), bs='cs')
    , family=binomial(link='logit')
    , data=train_xy
    , method='fREML'
  )
  y <- test_xy$EloCardinal
  pred <- predict(fitted_model, test_xy, type='response', se.fit=TRUE)
  cv_ts_plusDateRE[k] <- mean((y - as.vector(pred$fit))^2, na.rm=TRUE)
}
cv_ts_plusDateRE_avg <- mean(cv_ts_plusDateRE)
range(cv_ts_plusDateRE)
